# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from xlcompose import (
    DataFrame, Series, Row, Column, Tabs, CSpacer, RSpacer, Title, Image,
    VSpacer, HSpacer)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
from chainladder.utils.cupy import cp
from scipy.sparse import coo_matrix
import json
import joblib


class TriangleIO():
    def to_pickle(self, path, protocol=None):
        joblib.dump(self, filename=path, protocol=protocol)

    def to_json(self):
        ''' Serializes triangle object to json format

        Returns
        -------
            string representation of object in json format
        '''
        def sparse_out(tri):
            k, v, o, d = tri.shape
            xp = cp.get_array_module(tri)
            if xp == cp != np:
                out = cp.asnumpy(tri)
            else:
                out = tri
            coo = coo_matrix(np.nan_to_num(out.reshape((k*v*o, d))))
            return json.dumps(dict(zip([str(item) for item in zip(coo.row, coo.col)], coo.data)))

        json_dict = {}
        if self.is_val_tri:
            ddims = self.ddims
            json_dict['ddims'] = {
                'dtype': str(ddims.dtype),
                'array': ddims.values.tolist()}
            attributes = ['kdims', 'vdims', 'odims']
        else:
            attributes = ['kdims', 'vdims', 'odims', 'ddims']
        for attribute in attributes:
            json_dict[attribute] = {
                'dtype': str(getattr(self, attribute).dtype),
                'array': getattr(self, attribute).tolist()}
        xp = cp.get_array_module(self.values)
        if xp == cp != np:
            out = cp.asnumpy(self.cum_to_incr().values)
        else:
            out = self.cum_to_incr().values
        if np.sum(np.nan_to_num(out)==0) / np.prod(self.shape) > 0.40:
            json_dict['values'] = {
                'dtype': str(out.dtype),
                'array': sparse_out(out),
                'sparse': True}
        else:
            json_dict['values'] = {
                'dtype': str(out.dtype),
                'array': out.tolist(),
                'sparse': False}
        json_dict['key_labels'] = self.key_labels
        json_dict['origin_grain'] = self.origin_grain
        json_dict['development_grain'] = self.development_grain
        json_dict['nan_override'] = self.nan_override
        json_dict['is_cumulative'] = self.is_cumulative
        json_dict['is_val_tri'] = self.is_val_tri
        json_dict['valuation_date'] = self.valuation_date.strftime('%Y-%m-%d')
        return json.dumps(json_dict)


class EstimatorIO:
    ''' Class intended to allow persistence of estimator objects
    '''

    def to_pickle(self, path, protocol=None):
        joblib.dump(self, filename=path, protocol=protocol)

    def to_json(self):
        ''' Serializes triangle object to json format

        Returns
        -------
            string representation of object in json format
        '''
        return json.dumps(
            {'params': self.get_params(),
             '__class__': self.__class__.__name__})

    def __contains__(self, value):
        if self.__dict__.get(value, None) is None:
            return False
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np

from chainladder import ARRAY_BACKEND


try:
    import cupy as cp
    cp.array([1])
    module = 'cupy'
except:
    if ARRAY_BACKEND == 'cupy':
        import warnings
        warnings.warn('Unable to load CuPY.  Using numpy instead.')
    import numpy as cp
    module = 'numpy'

def get_array_module(*args, **kwargs):
    """ default array module when cupy is not present """
    return np

def nansum(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.sum(cp.nan_to_num(a), *args, **kwargs)

def nanmean(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.sum(cp.nan_to_num(a), *args, **kwargs) / \
           cp.sum(~cp.isnan(a), *args, **kwargs)

def nanmedian(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.array(np.nanmedian(cp.asnumpy(a), *args, **kwargs))

def nanpercentile(a, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.array(np.nanpercentile(cp.asnumpy(a), *args, **kwargs))

def unique(ar, axis=None, *args, **kwargs):
    """ For cupy v0.6.0 compatibility """
    return cp.array(np.unique(cp.asnumpy(ar), axis=axis, *args, **kwargs))

if module == 'cupy' and int(cp.__version__.split('.')[0]) < 7:
    cp.nansum = nansum
    cp.nanmean = nanmean
    cp.nanmedian = nanmedian
    cp.nanpercentile = nanpercentile
    cp.unique = unique

if module == 'numpy':
    cp.get_array_module = get_array_module



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
from chainladder.utils.cupy import cp
import pandas as pd
import copy
import warnings
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder import WeightedRegression
from chainladder.core import EstimatorIO


class DevelopmentBase(BaseEstimator, TransformerMixin, EstimatorIO):
    @staticmethod
    def _get_cdf(obj):
        if 'ldf_' not in obj:
            return
        else:
            obj2 = copy.copy(obj.ldf_)
            xp = cp.get_array_module(obj2.values)
            cdf_ = xp.flip(xp.cumprod(xp.flip(obj2.values, -1), -1), -1)
            obj2.ddims = xp.array(
                [item.replace(item[item.find("-")+1:], '9999')
                 for item in obj2.ddims])
            obj2.values = cdf_
            obj2._set_slicers()
            return obj2


class Development(DevelopmentBase):
    """ A Transformer that allows for basic loss development pattern selection.

    Parameters
    ----------
    n_periods : integer, optional (default=-1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods=-1
    average : string, optional (default='volume')
        type of averaging to use for ldf average calculation.  Options include
        'volume', 'simple', and 'regression'
    sigma_interpolation : string optional (default='log-linear')
        Options include 'log-linear' and 'mack'
    drop : tuple or list of tuples
        Drops specific origin/development combination(s)
    drop_high : bool or list of bool (default=None)
        Drops highest link ratio(s) from LDF calculation
    drop_low : bool or list of bool (default=None)
        Drops lowest link ratio(s) from LDF calculation
    drop_valuation : str or list of str (default=None)
        Drops specific valuation periods. str must be date convertible.
    fillna: float, (default=None)
        Used to fill in zero or nan values of an triangle with some non-zero
        amount.  When an link-ratio has zero as its denominator, it is automatically
        excluded from the `ldf_` calculation.  For the specific case of 'volume'
        averaging in a deterministic method, this may be reasonable.  For all other
        averages and stochastic methods, this assumption should be avoided.


    Attributes
    ----------
    ldf_ : Triangle
        The estimated loss development patterns
    cdf_ : Triangle
        The estimated cumulative development patterns
    sigma_ : Triangle
        Sigma of the ldf regression
    std_err_ : Triangle
        Std_err of the ldf regression
    weight_ : pandas.DataFrame
        The weight used in the ldf regression

    """
    def __init__(self, n_periods=-1, average='volume',
                 sigma_interpolation='log-linear', drop=None,
                 drop_high=None, drop_low=None, drop_valuation=None,
                 fillna=None):
        self.n_periods = n_periods
        self.average = average
        self.sigma_interpolation = sigma_interpolation
        self.drop_high = drop_high
        self.drop_low = drop_low
        self.drop_valuation = drop_valuation
        self.drop = drop
        self.fillna = fillna

    @property
    def weight_(self):
        return pd.DataFrame(
            self.w_[0, 0], index=self.ldf_.origin,
            columns=list(self.ldf_.development.values[:,0]))

    def _assign_n_periods_weight(self, X):
        if type(self.n_periods) is int:
            return self._assign_n_periods_weight_int(X, self.n_periods)[..., :-1]
        elif type(self.n_periods) is list:
            if len(self.n_periods) != X.values.shape[-1]-1:
                raise ValueError('n_periods list must be of lenth {}.'
                                 .format(X.values.shape[-1]-1))
            else:
                return self._assign_n_periods_weight_list(X)
        else:
            raise ValueError('n_periods must be of type <int> or <list>')

    def _assign_n_periods_weight_list(self, X):
        ''' private method to standardize the n_periods input to a list '''
        xp = cp.get_array_module(X.values)
        dict_map = {item: self._assign_n_periods_weight_int(X, item)
                    for item in set(self.n_periods)}
        conc = [dict_map[item][..., num:num+1, :]
                for num, item in enumerate(self.n_periods)]
        return xp.swapaxes(xp.concatenate(tuple(conc), -2), -2, -1)

    def _assign_n_periods_weight_int(self, X, n_periods):
        ''' Zeros out weights depending on number of periods desired
            Only works for type(n_periods) == int
        '''
        xp = cp.get_array_module(X.values)
        if n_periods < 1 or n_periods >= X.shape[-2] - 1:
            return X.values*0+1
        else:
            val_offset = {
                'Y': {'Y': 1},
                'Q': {'Y':4, 'Q': 1},
                'M': {'Y':12, 'Q': 3, 'M': 1}}
            val_date_min = \
                X.valuation[X.valuation<=X.valuation_date].drop_duplicates().sort_values()
            val_date_min = \
                val_date_min[-n_periods * \
                val_offset[X.development_grain][X.origin_grain] - 1]
            w = X[X.valuation>=val_date_min]
            return xp.nan_to_num((w/w).values)*X._expand_dims(X._nan_triangle())


    def _drop_adjustment(self, X, link_ratio):
        weight = X._nan_triangle()[:, :-1]
        if self.drop_high == self.drop_low == \
           self.drop == self.drop_valuation is None:
            return weight
        if self.drop_high is not None:
            weight = weight*self._drop_hilo('high', X, link_ratio)
        if self.drop_low is not None:
            weight = weight*self._drop_hilo('low', X, link_ratio)
        if self.drop is not None:
            weight = weight*self._drop(X)
        if self.drop_valuation is not None:
            weight = weight*self._drop_valuation(X)
        return weight

    def _drop_hilo(self, kind, X, link_ratio):
        xp = cp.get_array_module(X.values)
        link_ratio[link_ratio == 0] = xp.nan
        link_ratio = link_ratio + np.random.rand(*list(link_ratio.shape))/1e8
        lr_valid_count = xp.sum(~xp.isnan(link_ratio)[0, 0], axis=0)
        if kind == 'high':
            vals = xp.nanmax(link_ratio, -2, keepdims=True)
            drop_hilo = self.drop_high
        else:
            vals = xp.nanmin(link_ratio, -2, keepdims=True)
            drop_hilo = self.drop_low
        hilo = 1*(vals != link_ratio)
        if type(drop_hilo) is bool:
            drop_hilo = [drop_hilo]*(len(X.development)-1)
        for num, item in enumerate(self.average_):
            if not drop_hilo[num]:
                hilo[..., num] = hilo[..., num]*0+1
            else:
                if lr_valid_count[num] < 3:
                    hilo[..., num] = hilo[..., num]*0+1
                    warnings.warn('drop_high and drop_low cannot be computed '
                                  'when less than three LDFs are present. '
                                  'Ignoring exclusions in some cases.')
        return hilo

    def _drop_valuation(self, X):
        xp = cp.get_array_module(X.values)
        if type(self.drop_valuation) is not list:
            drop_valuation = [self.drop_valuation]
        else:
            drop_valuation = self.drop_valuation
        arr = 1-xp.nan_to_num(X[X.valuation.isin(
            pd.PeriodIndex(drop_valuation,
                           freq=X.origin_grain).to_timestamp(how='e'))].values[0, 0]*0+1)
        ofill = X.shape[-2]-arr.shape[-2]
        dfill = X.shape[-1]-arr.shape[-1]
        if ofill > 0:
            arr = xp.concatenate((arr, xp.repeat(
                xp.ones(arr.shape[-1])[xp.newaxis], ofill, 0)), 0)
        if dfill > 0:
            arr = xp.concatenate((arr, xp.repeat(
                xp.ones(arr.shape[-2])[..., xp.newaxis], dfill, -1)), -1)
        return arr[:, :-1]

    def _drop(self, X):
        xp = cp.get_array_module(X.values)
        drop = [self.drop] if type(self.drop) is not list else self.drop
        arr = X._nan_triangle().copy()
        for item in drop:
            arr[np.where(X.origin == item[0])[0][0],
                np.where(X.development == item[1])[0][0]] = 0
        return arr[:, :-1]

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        xp = cp.get_array_module(X.values)
        if (type(X.ddims) != np.ndarray):
            raise ValueError('Triangle must be expressed with development lags')
        if self.fillna:
            tri_array = (X + self.fillna).values
        else:
            tri_array = X.values.copy()
        tri_array[tri_array == 0] = xp.nan
        if type(self.average) is str:
            average = [self.average] * (tri_array.shape[-1] - 1)
        else:
            average = self.average
        average = np.array(average)
        self.average_ = average
        weight_dict = {'regression': 0, 'volume': 1, 'simple': 2}
        x, y = tri_array[..., :-1], tri_array[..., 1:]
        val = xp.array([weight_dict.get(item.lower(), 1)
                        for item in average])
        for i in [2, 1, 0]:
            val = xp.repeat(val[xp.newaxis], tri_array.shape[i], axis=0)
        val = xp.nan_to_num(val * (y * 0 + 1))
        if xp == cp:
            link_ratio = y / x
        else:
            link_ratio = xp.divide(y, x, where=xp.nan_to_num(x) != 0)
        self.w_ = xp.array(self._assign_n_periods_weight(X) *
                           self._drop_adjustment(X, link_ratio),
                           dtype='float16')
        w = self.w_ / (x**(val))
        params = WeightedRegression(axis=2, thru_orig=True).fit(x, y, w)
        if self.n_periods != 1:
            params = params.sigma_fill(self.sigma_interpolation)
        else:
            warnings.warn('Setting n_periods=1 does not allow enough degrees '
                          'of freedom to support calculation of all regression'
                          ' statistics.  Only LDFs have been calculated.')
        params.std_err_ = xp.nan_to_num(params.std_err_) + \
            xp.nan_to_num(
                (1-xp.nan_to_num(params.std_err_*0+1)) *
                params.sigma_ /
                xp.swapaxes(xp.sqrt(x**(2-val))[..., 0:1, :], -1, -2))
        params = xp.concatenate(
            (params.slope_, params.sigma_, params.std_err_), 3)
        params = xp.swapaxes(params, 2, 3)
        self.ldf_ = self._param_property(X, params, 0)
        self.cdf_ = self._get_cdf(self)
        self.sigma_ = self._param_property(X, params, 1)
        self.std_err_ = self._param_property(X, params, 2)
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = copy.copy(X)
        triangles = ['std_err_', 'cdf_', 'ldf_', 'sigma_']
        for item in triangles + ['average_', 'w_', 'sigma_interpolation']:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new

    def _param_property(self, X, params, idx):
        obj = copy.copy(X)
        xp = cp.get_array_module(X.values)
        obj.values = xp.ones(X.shape)[..., :-1]*params[..., idx:idx+1, :]
        obj.ddims = X.link_ratio.ddims
        obj.valuation = obj._valuation_triangle(obj.ddims)
        obj.nan_override = True
        obj._set_slicers()
        return obj



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import copy
import numpy as np
from chainladder.utils.cupy import cp
from chainladder.utils import WeightedRegression
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.development import DevelopmentBase, Development
from chainladder.core import EstimatorIO


class TailBase(BaseEstimator, TransformerMixin, EstimatorIO):
    ''' Base class for all tail methods.  Tail objects are equivalent
        to development objects with an additional set of tail statistics'''
    def fit(self, X, y=None, sample_weight=None):
        obj = copy.copy(X)
        xp = cp.get_array_module(obj.values)
        if 'ldf_'not in obj:
            obj = Development().fit_transform(obj)
        self._ave_period = {'Y': (1, 12),
                            'Q': (4, 3),
                            'M': (12, 1)}[obj.development_grain]
        ddims = np.concatenate(
            (obj.ddims, [(item+1)*self._ave_period[1] + obj.ddims[-1]
                         for item in range(self._ave_period[0])], [9999]), 0)
        self.ldf_ = copy.copy(obj.ldf_)
        tail = xp.ones(self.ldf_.shape)[..., -1:]
        tail = xp.repeat(tail, self._ave_period[0]+1, -1)
        self.ldf_.values = xp.concatenate((self.ldf_.values, tail), -1)
        self.ldf_.ddims = np.array(['{}-{}'.format(ddims[i], ddims[i+1])
                                    for i in range(len(ddims)-1)])
        self.ldf_.valuation = self.ldf_._valuation_triangle()
        self.sigma_ = copy.copy(getattr(obj, 'sigma_', obj.cdf_*0))
        self.std_err_ = copy.copy(getattr(obj, 'std_err_', obj.cdf_*0))
        zeros = tail[..., -1:]*0
        self.sigma_.values = xp.concatenate(
            (self.sigma_.values, zeros), -1)
        self.std_err_.values = xp.concatenate(
            (self.std_err_.values, zeros), -1)
        self.sigma_.ddims = self.std_err_.ddims = \
            np.concatenate(
                (obj.ldf_.ddims,
                 np.array(['{}-9999'.format(int(obj.ddims[-1]))])))
        val_array = self.sigma_._valuation_triangle(self.sigma_.ddims)
        self.sigma_.valuation = self.std_err_.valuation = val_array
        self.cdf_ = DevelopmentBase._get_cdf(self)
        self.cdf_._set_slicers()
        self.ldf_._set_slicers()
        self.sigma_._set_slicers()
        self.std_err_._set_slicers()
        return self

    def transform(self, X):
        X_new = copy.deepcopy(X)
        xp = cp.get_array_module(X.values)
        X_new.std_err_.values = xp.concatenate(
            (X_new.std_err_.values,
             self.std_err_.values[..., -1:]), -1)
        X_new.cdf_.values = xp.concatenate(
            (X_new.cdf_.values,
             self.cdf_.values[..., -self._ave_period[0]-1:]*0+1), -1)
        X_new.cdf_.values = X_new.cdf_.values * \
            self.cdf_.values[..., -self._ave_period[0]-1:-self._ave_period[0]]
        X_new.cdf_.values[..., -1] = self.cdf_.values[..., -1]
        X_new.ldf_.values = xp.concatenate(
            (X_new.ldf_.values,
             self.ldf_.values[..., -self._ave_period[0]-1:]), -1)
        X_new.sigma_.values = xp.concatenate(
            (X_new.sigma_.values, self.sigma_.values[..., -1:]), -1)
        X_new.cdf_.ddims = X_new.ldf_.ddims = self.ldf_.ddims
        X_new.sigma_.ddims = X_new.std_err_.ddims = self.sigma_.ddims
        X_new.cdf_.valuation = X_new.ldf_.valuation = self.ldf_.valuation
        X_new.sigma_.valuation = \
            X_new.std_err_.valuation = self.sigma_.valuation
        X_new.sigma_._set_slicers()
        X_new.ldf_._set_slicers()
        X_new.cdf_._set_slicers()
        X_new.std_err_._set_slicers()
        return X_new

    def fit_transform(self, X, y=None, sample_weight=None):
        """ Equivalent to fit(X).transform(X)

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs based on the model.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        self.fit(X)
        if 'std_err_' not in X:
            X = Development().fit_transform(X)
        return self.transform(X)

    def _get_tail_prediction(self, tail_ldf):
        xp = cp.get_array_module(tail_ldf)
        accum_point = self.ldf_.shape[-1] - 1
        ave = 1 + tail_ldf[..., :accum_point]
        all = xp.expand_dims(xp.prod(1 + tail_ldf[..., accum_point:], -1), -1)
        tail = xp.concatenate((ave, all), -1)
        return tail

    def _get_initial_ldf(self, xp, tail):
        ''' Quadratic series expansion solution to return seed LDF for tail'''
        arr = self.decay**xp.arange(1000)
        a = xp.sum(arr**2)
        b = xp.sum(arr)
        c = -xp.log(tail)
        return (-b+xp.sqrt(b**2-4*a*c))/(2*a)

    def _apply_decay(self, X, tail):
        ''' Created Tail vector with decay over time '''
        xp = cp.get_array_module(X.values)
        decay_range = self.ldf_.shape[-1]-X.shape[-1]+1
        if xp.max(tail) == 1.0:
            ldfs = 1*(self.decay**xp.arange(1000))
        else:
            ldfs = 1+self._get_initial_ldf(xp, tail)*(self.decay**xp.arange(1000))
        ldfs = ldfs[..., :decay_range]
        ldfs[..., -1:] = tail/xp.prod(ldfs[..., :-1], axis=-1, keepdims=True)
        self.ldf_.values[..., -decay_range:] = \
            self.ldf_.values[..., -decay_range:]*ldfs
        self.cdf_ = DevelopmentBase._get_cdf(self)
        return self

    def _get_tail_stats(self, X):
        """ Method to approximate the tail sigma using
        log-linear extrapolation applied to tail average period
        """
        time_pd = self._get_tail_weighted_time_period(X)
        xp = cp.get_array_module(X.sigma_.values)
        reg = WeightedRegression(axis=3).fit(None, xp.log(X.sigma_.values), None)
        sigma_ = xp.exp(time_pd*reg.slope_+reg.intercept_)
        y = X.std_err_.values
        y[y == 0] = xp.nan
        reg = WeightedRegression(axis=3).fit(None, xp.log(y), None)
        std_err_ = xp.exp(time_pd*reg.slope_+reg.intercept_)
        return sigma_, std_err_

    def _get_tail_weighted_time_period(self, X):
        """ Method to approximate the weighted-average development age of tail
        using log-linear extrapolation

        Returns: float32
        """
        y = X.ldf_.values.copy()
        xp = cp.get_array_module(y)
        y[y <= 1] = xp.nan
        reg = WeightedRegression(axis=3).fit(None, xp.log(y - 1), None)
        tail = xp.prod(self.ldf_.values[..., -self._ave_period[0]-1:],
                       -1, keepdims=True)
        reg = WeightedRegression(axis=3).fit(None, xp.log(y - 1), None)
        time_pd = (xp.log(tail-1)-reg.intercept_)/reg.slope_
        return time_pd



ARRAY_BACKEND = 'numpy'

def array_backend(array_backend='numpy'):
    global ARRAY_BACKEND
    ARRAY_BACKEND = array_backend

from chainladder.utils import * # noqa (API Import)
from chainladder.core import * # noqa (API Import)
from chainladder.development import * # noqa (API Import)
from chainladder.tails import * # noqa (API Import)
from chainladder.methods import * # noqa (API Import)
from chainladder.workflow import * # noqa (API Import)

__version__ = '0.6.0'



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from chainladder.methods.chainladder import Chainladder
from chainladder.development.base import DevelopmentBase, Development
import numpy as np
from chainladder.utils.cupy import cp
import pandas as pd
import copy
from warnings import warn
import types


class BootstrapODPSample(DevelopmentBase):
    """
    Class to generate bootstrap samples of triangles.  Currently this Only
    supports 'single' triangles (single index and single column).

    Parameters
    ----------
    n_sims : int (default=1000)
        Number of simulations to generate
    n_periods : integer, optional (default=-1)
        number of origin periods to be used in the ldf average calculation. For
        all origin periods, set n_periods=-1
    hat_adj : bool (default=False)
        Adjust standardized Pearson residuals with the hat matrix adjustment
        factor.  If false, Degree of Freedom adjustment is used.
    drop : tuple or list of tuples
        Drops specific origin/development combination(s) from residual sample
    drop_high : bool or list of bool (default=None)
        Drops highest link ratio(s) from residual sample
    drop_low : bool or list of bool (default=None)
        Drops lowest link ratio(s) from residual sample
    drop_valuation : str or list of str (default=None)
        Drops specific valuation periods from residual sample. str must be date
        convertible.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random.

    Attributes
    ----------
    resampled_triangles_ : Triangle
        A set of triangles represented by each simulation
    scale_ :
        The scale parameter to be used in generating process risk
    """
    def __init__(self, n_sims=1000, n_periods=-1,
                 hat_adj=True, drop=None, drop_high=None, drop_low=None,
                 drop_valuation=None, random_state=None):
        self.n_sims = n_sims
        self.n_periods = n_periods
        self.hat_adj = hat_adj
        self.drop = drop
        self.drop_high = drop_high
        self.drop_low = drop_low
        self.drop_valuation = drop_valuation
        self.random_state = random_state

    def fit(self, X, y=None, sample_weight=None):
        xp = cp.get_array_module(X.values)
        if (type(X.ddims) != np.ndarray):
            raise ValueError('Triangle must be expressed with development lags')
        obj = copy.copy(X)
        lag = {'M': 1, 'Q': 3, 'Y': 12}[X.development_grain]
        obj = Development(
            n_periods=self.n_periods, drop=self.drop,
            drop_high=self.drop_high, drop_low=self.drop_low,
            drop_valuation=self.drop_valuation).fit_transform(obj)
        self.w_ = obj.w_
        obj = Chainladder().fit(obj)
        # Works for only a single triangle - can we generalize this
        exp_incr_triangle = obj.full_expectation_.cum_to_incr() \
                               .values[0, 0, :, :X.shape[-1]]
        exp_incr_triangle = xp.nan_to_num(exp_incr_triangle) * \
            obj.X_._nan_triangle()
        self.design_matrix_ = self._get_design_matrix(X)
        if self.hat_adj:
            try:
                self.hat_ = self._get_hat(X, exp_incr_triangle)
            except:
                warn('Could not compute hat matrix.  Setting hat_adj to False')
                self.had_adj = False
                self.hat_ = None
        else:
            self.hat_ = None
        self.resampled_triangles_, self.scale_ = \
            self._get_simulation(X, exp_incr_triangle)
        n_obs = xp.nansum(self.w_)
        n_origin_params = X.shape[2]
        n_dev_params = X.shape[3] - 1
        deg_free = n_obs - n_origin_params - n_dev_params
        deg_free_adj_fctr = np.sqrt(n_obs/deg_free)
        return self

    def _get_simulation(self, X, exp_incr_triangle):
        xp = cp.get_array_module(X.values)
        k_value = 1  # for ODP Poisson
        if X.shape[:2] != (1, 1):
            raise ValueError('Only single index/column triangles are ',
                             'supported')
        unscaled_residuals = \
            ((X.cum_to_incr().values - exp_incr_triangle) /
             xp.sqrt(xp.abs(exp_incr_triangle**k_value)))[0, 0, ...]
        w_ = self.w_[0, 0]
        w_[:, 1:]*w_[:, 1:]
        w_ = xp.concatenate((w_[:, 0:1], w_), axis=1)
        unscaled_residuals = unscaled_residuals * w_
        pearson_chi_sq = sum(sum(xp.nan_to_num(unscaled_residuals)**2))
        if self.hat_ is None:
            standardized_residuals = unscaled_residuals
        else:
            standardized_residuals = self.hat_ * unscaled_residuals
        n_params = self.design_matrix_.shape[1]
        degree_freedom = xp.nansum(X._nan_triangle()) - n_params
        # Shapland has a hetero adjustment to degree_freedom here
        # He also adjusts the residuals for the hetero adjustment
        scale_phi = pearson_chi_sq / degree_freedom
        k, v, o, d = X.shape
        resids = xp.reshape(standardized_residuals, (k, v, o*d))

        adj_resid_dist = resids[xp.isfinite(resids)]  # Missing k,v dimensions
        # Suggestions from Using the ODP Bootstrap Model: A Practitioners Guide
        adj_resid_dist = adj_resid_dist[adj_resid_dist != 0]
        adj_resid_dist = adj_resid_dist - xp.mean(adj_resid_dist)
        random_state = xp.random.RandomState(self.random_state)
        resampled_residual = [xp.expand_dims(random_state.choice(adj_resid_dist,
                              size=exp_incr_triangle.shape,
                              replace=True)*(exp_incr_triangle*0+1), 0)
                              for item in range(self.n_sims)]
        resampled_residual = xp.concatenate(tuple(resampled_residual), 0) \
                               .reshape(self.n_sims, exp_incr_triangle.shape[0],
                                        exp_incr_triangle.shape[1])
        resampled_residual = resampled_residual
        b = xp.repeat(xp.expand_dims(exp_incr_triangle, 0), self.n_sims, 0)
        resampled_triangles = (resampled_residual*xp.sqrt(abs(b))+b).cumsum(2)
        resampled_triangles = xp.swapaxes(
            xp.expand_dims(resampled_triangles, 0), 0, 1)
        obj = copy.copy(X)
        obj.kdims = np.arange(self.n_sims)
        obj.values = resampled_triangles
        obj._set_slicers()
        return obj, scale_phi

    def _get_design_matrix(self, X):
        """ The design matrix used in hat matrix adjustment (Shapland eq3.12)
        """
        xp = cp.get_array_module(X.values)
        w = X._nan_triangle()
        arr = xp.diag(w[:, 0])
        intra_beta = xp.zeros((w.shape[0], w.shape[1]-1))
        arr = xp.concatenate((arr, intra_beta), axis=1)
        for i in range(w.shape[1]-1):
            len_alpha = len(w[:, i+1][~xp.isnan(w[:, i+1])])
            intra_alpha = xp.diag(w[:, i+1])[:len_alpha, :]
            intra_beta[:, i] = 1
            intra_beta = intra_beta[:len_alpha, :]
            intra_arr = xp.concatenate((intra_alpha, intra_beta), axis=1)
            arr = xp.concatenate((arr, intra_arr), axis=0)
        return arr

    def _get_hat(self, X, exp_incr_triangle):
        """ The hat matrix adjustment (Shapland eq3.23)"""
        xp = cp.get_array_module(X.values)
        weight_matrix = xp.diag(pd.DataFrame(exp_incr_triangle).unstack().dropna().values)
        design_matrix = self.design_matrix_
        hat = xp.matmul(xp.matmul(xp.matmul(design_matrix,xp.linalg.inv(xp.matmul(design_matrix.T, xp.matmul(weight_matrix, design_matrix)))), design_matrix.T), weight_matrix)
        hat = xp.diagonal(xp.sqrt(xp.divide(1, abs(1-hat), where=(1-hat)!=0)))
        total_length = X._nan_triangle().shape[0]
        reshaped_hat = xp.reshape(hat[:total_length],(1, total_length))
        indices = xp.nansum(X._nan_triangle(), axis=0).cumsum().astype(int)
        for num, item in enumerate(indices[:-1]):
            col_length = int(indices[num+1]-indices[num])
            col = xp.reshape(hat[int(indices[num]):int(indices[num+1])],(1,col_length))
            nans = xp.repeat(xp.expand_dims(xp.array([xp.nan]),0),total_length-col_length, axis=1)
            col = xp.concatenate((col, nans), axis=1)
            reshaped_hat = xp.concatenate((reshaped_hat,col),axis=0)
        return reshaped_hat.T

    def _get_hetero_adjustment(self):
        pass

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = copy.copy(X)
        X_new = self.resampled_triangles_
        X_new.scale_ = self.scale_
        X_new.random_state = self.random_state
        X_new._get_process_variance = types.MethodType(_get_process_variance, X_new)
        return X_new

def _get_process_variance(self, full_triangle):
    #if self.process_dist == 'od poisson':
    #    process_triangle = np.nan_to_num(np.array([random_state.poisson(lam=abs(item))*np.sign(np.nan_to_num(item))for item in sim_exp_incr_triangle]))
    xp = cp.get_array_module(full_triangle.values)
    lower_tri = full_triangle.cum_to_incr() - self.cum_to_incr()
    random_state = xp.random.RandomState(None if not self.random_state else self.random_state + 1)
    lower_tri.values = random_state.gamma(
        shape=abs(lower_tri.values) / self.scale_, scale=self.scale_) * \
        np.sign(np.nan_to_num(lower_tri.values))
    return (lower_tri + self.cum_to_incr()).incr_to_cum()



import chainladder as cl
raa = cl.load_dataset('RAA')
raa_1989 = raa[raa.valuation < raa.valuation_date]
cl_ult = cl.Chainladder().fit(raa).ultimate_  # Chainladder Ultimate
apriori = cl_ult*0+(float(cl_ult.sum())/10)  # Mean Chainladder Ultimate
apriori_1989 = apriori[apriori.origin < '1990']


def test_cc_predict():
    cc = cl.CapeCod().fit(raa_1989, sample_weight=apriori_1989)
    cc.predict(raa, sample_weight=apriori)


def test_bf_predict():
    cc = cl.BornhuetterFerguson().fit(raa_1989, sample_weight=apriori_1989)
    cc.predict(raa, sample_weight=apriori)


def test_mack_predict():
    mack = cl.MackChainladder().fit(raa_1989)
    mack.predict(raa)

def test_bs_random_state_predict():
    tri = cl.load_dataset('clrd').groupby('LOB').sum().loc['wkcomp', ['CumPaidLoss', 'EarnedPremNet']]
    X = cl.BootstrapODPSample(random_state=100).fit_transform(tri['CumPaidLoss'])
    bf = cl.BornhuetterFerguson(apriori=0.6, apriori_sigma=0.1, random_state=42).fit(X, sample_weight=tri['EarnedPremNet'].latest_diagonal)
    assert bf.predict(X, sample_weight=tri['EarnedPremNet'].latest_diagonal).ibnr_ == bf.ibnr_



import chainladder as cl

def test_bondy1():
    tri = cl.load_dataset('tail_sample')['paid']
    dev = cl.Development(average='simple').fit_transform(tri)
    assert round(cl.TailBondy().fit(dev).cdf_.values[0,0,0,-2],3) == 1.045



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
from scipy.optimize import least_squares
from chainladder.utils.cupy import cp
from chainladder.tails import TailBase
from chainladder.development import DevelopmentBase, Development

class TailBondy(TailBase):
    """Estimator for the Generalized Bondy tail factor.

    .. versionadded:: 0.6.0

    Parameters
    ----------
    earliest_age : int
        The earliest age from which the Bondy exponent is to be calculated.
        Defaults to earliest available in the Triangle.
    decay : float (default=0.50)
        An exponential decay constant that allows for decay over future
        development periods.  A decay rate of 0.5 sets the development portion
        of each successive LDF to 50% of the previous LDF.

    Attributes
    ----------
    b_ :
        The Bondy exponent
    ldf_ :
        ldf with tail applied.
    cdf_ :
        cdf with tail applied.
    sigma_ :
        sigma with tail factor applied.
    std_err_ :
        std_err with tail factor applied

    See also
    --------
    TailCurve

    """
    def __init__(self, earliest_age=None, decay=0.5):
        self.earliest_age = earliest_age
        self.decay = decay

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the tail will be applied.
        y : Ignored
        sample_weight : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        super().fit(X, y, sample_weight)
        xp = cp.get_array_module(X.values)
        obj = Development().fit_transform(X) if 'ldf_' not in X else X
        b_optimized = []
        initial = xp.where(obj.ddims==self.earliest_age)[0][0] if self.earliest_age else 0
        for num in range(len(obj.vdims)):
            b0 = xp.ones(obj.shape[0])*.5
            data = xp.log(obj.ldf_.values[:, num, 0,initial:])
            b_optimized.append(least_squares(
                TailBondy.solver, x0=b0, kwargs={'data': data}).x[:, None])
        self.b_ = xp.concatenate(b_optimized, 1)
        tail = xp.exp(xp.log(obj.ldf_.values[..., 0:1, initial:initial+1]) * \
                      self.b_**(len(obj.ldf_.ddims)-1))
        tail = (tail**(self.b_/(1-self.b_)))*tail
        self = self._apply_decay(obj, tail)
        sigma, std_err = self._get_tail_stats(obj)
        self.sigma_.values[..., -1] = sigma[..., -1]
        self.std_err_.values[..., -1] = std_err[..., -1]
        return self

    def transform(self, X):
        """Transform X.
        The predicted class of an input sample is a vote by the trees in
        the forest, weighted by their probability estimates. That is,
        the predicted class is the one with highest mean probability
        estimate across the trees.

        Parameters
        ----------
        X : Triangle
            Triangle must contain the ``ldf_`` development attribute.

        Returns
        -------
        X_new : Triangle
            New Triangle with tail factor applied to its development
            attributes.
        """
        X.std_err_ = self.std_err_
        X.cdf_ = self.cdf_
        X.ldf_ = self.ldf_
        X.sigma_ = self.sigma_
        X.b_ = self.b_
        return X

    @staticmethod
    def solver(b, data):
        xp = cp.get_array_module(data)
        arange = xp.repeat(xp.arange(data.shape[-1])[None, :], data.shape[0], 0)
        out = data - (data[:, 0])[:, None]*b**(arange)
        return out.flatten()



import chainladder as cl
import numpy as np
from chainladder.utils.cupy import cp


def test_constant_balances():
    raa = cl.load_dataset('quarterly')
    xp = cp.get_array_module(raa.values)
    assert round(float(xp.prod(cl.TailConstant(1.05, decay=0.8)
                     .fit(raa).ldf_.iloc[0, 1].values[0, 0, 0, -5:])),3) == 1.050
